@@ -18,12 +18,23 @@ This module contains builtin handlers for events emitted by botocore.
 
 import base64
 import hashlib
-import functools
 import logging
-import re
 import xml.etree.cElementTree
+import copy
+import re
+import warnings
+
+from botocore.compat import urlsplit, urlunsplit, unquote, \
+    json, quote, six, unquote_str
+from botocore.docs.utils import AutoPopulatedParam
+from botocore.docs.utils import HideParamFromOperations
+from botocore.docs.utils import AppendParamDocumentation
+from botocore.signers import add_generate_presigned_url
+from botocore.signers import add_generate_presigned_post
+from botocore.exceptions import ParamValidationError
+from botocore.exceptions import UnsupportedTLSVersionWarning
+from botocore.utils import percent_encode, SAFE_CHARS
 
-from botocore.compat import urlsplit, urlunsplit, unquote, json, quote, six
 from botocore import retryhandler
 from botocore import utils
 from botocore import translate
@@ -32,14 +43,16 @@ import botocore.auth
 
 
 logger = logging.getLogger(__name__)
-LABEL_RE = re.compile('[a-z0-9][a-z0-9\-]*[a-z0-9]')
-RESTRICTED_REGIONS = [
-    'us-gov-west-1',
-    'fips-us-gov-west-1',
-]
+
 REGISTER_FIRST = object()
 REGISTER_LAST = object()
-
+# From the S3 docs:
+# The rules for bucket names in the US Standard region allow bucket names
+# to be as long as 255 characters, and bucket names can contain any
+# combination of uppercase letters, lowercase letters, numbers, periods
+# (.), hyphens (-), and underscores (_).
+VALID_BUCKET = re.compile('^[a-zA-Z0-9.\-_]{1,255}$')
+VERSION_ID_SUFFIX = re.compile(r'\?versionId=[^\s]+$')
 
 
 def check_for_200_error(response, **kwargs):
@@ -65,8 +78,8 @@ def check_for_200_error(response, **kwargs):
     http_response, parsed = response
     if _looks_like_special_case_error(http_response):
         logger.debug("Error found for response with 200 status code, "
-                        "errors: %s, changing status code to "
-                        "500.", parsed)
+                     "errors: %s, changing status code to "
+                     "500.", parsed)
         http_response.status_code = 500
 
 
@@ -83,11 +96,12 @@ def _looks_like_special_case_error(http_response):
 
 
 def decode_console_output(parsed, **kwargs):
-    try:
-        value = base64.b64decode(six.b(parsed['Output'])).decode('utf-8')
-        parsed['Output'] = value
-    except (ValueError, TypeError, AttributeError):
-        logger.debug('Error decoding base64', exc_info=True)
+    if 'Output' in parsed:
+        try:
+            value = base64.b64decode(six.b(parsed['Output'])).decode('utf-8')
+            parsed['Output'] = value
+        except (ValueError, TypeError, AttributeError):
+            logger.debug('Error decoding base64', exc_info=True)
 
 
 def decode_quoted_jsondoc(value):
@@ -109,11 +123,46 @@ def json_decode_template_body(parsed, **kwargs):
 
 def calculate_md5(params, **kwargs):
     request_dict = params
-    if request_dict['body'] and not 'Content-MD5' in params['headers']:
-        md5 = hashlib.md5()
-        md5.update(six.b(params['body']))
-        value = base64.b64encode(md5.digest()).decode('utf-8')
-        params['headers']['Content-MD5'] = value
+    if request_dict['body'] and 'Content-MD5' not in params['headers']:
+        body = request_dict['body']
+        if isinstance(body, bytes):
+            binary_md5 = _calculate_md5_from_bytes(body)
+        else:
+            binary_md5 = _calculate_md5_from_file(body)
+        base64_md5 = base64.b64encode(binary_md5).decode('ascii')
+        params['headers']['Content-MD5'] = base64_md5
+
+
+def _calculate_md5_from_bytes(body_bytes):
+    md5 = hashlib.md5(body_bytes)
+    return md5.digest()
+
+
+def _calculate_md5_from_file(fileobj):
+    start_position = fileobj.tell()
+    md5 = hashlib.md5()
+    for chunk in iter(lambda: fileobj.read(1024 * 1024), b''):
+        md5.update(chunk)
+    fileobj.seek(start_position)
+    return md5.digest()
+
+
+def conditionally_calculate_md5(params, **kwargs):
+    """Only add a Content-MD5 when not using sigv4"""
+    signer = kwargs['request_signer']
+    if signer.signature_version != 'v4':
+        calculate_md5(params, **kwargs)
+
+
+def validate_bucket_name(params, **kwargs):
+    if 'Bucket' not in params:
+        return
+    bucket = params['Bucket']
+    if VALID_BUCKET.search(bucket) is None:
+        error_msg = (
+            'Invalid bucket name "%s": Bucket name must match '
+            'the regex "%s"' % (bucket, VALID_BUCKET.pattern))
+        raise ParamValidationError(report=error_msg)
 
 
 def sse_md5(params, **kwargs):
@@ -123,105 +172,37 @@ def sse_md5(params, **kwargs):
     encryption key. This handler does both if the MD5 has not been set by
     the caller.
     """
-    if not _needs_s3_sse_customization(params):
+    _sse_md5(params, 'SSECustomer')
+
+
+def copy_source_sse_md5(params, **kwargs):
+    """
+    S3 server-side encryption requires the encryption key to be sent to the
+    server base64 encoded, as well as a base64-encoded MD5 hash of the
+    encryption key. This handler does both if the MD5 has not been set by
+    the caller specifically if the parameter is for the copy-source sse-c key.
+    """
+    _sse_md5(params, 'CopySourceSSECustomer')
+
+
+def _sse_md5(params, sse_member_prefix='SSECustomer'):
+    if not _needs_s3_sse_customization(params, sse_member_prefix):
         return
-    key_as_bytes = params['SSECustomerKey']
+    sse_key_member = sse_member_prefix + 'Key'
+    sse_md5_member = sse_member_prefix + 'KeyMD5'
+    key_as_bytes = params[sse_key_member]
     if isinstance(key_as_bytes, six.text_type):
         key_as_bytes = key_as_bytes.encode('utf-8')
     key_md5_str = base64.b64encode(
         hashlib.md5(key_as_bytes).digest()).decode('utf-8')
     key_b64_encoded = base64.b64encode(key_as_bytes).decode('utf-8')
-    params['SSECustomerKey'] = key_b64_encoded
-    params['SSECustomerKeyMD5'] = key_md5_str
+    params[sse_key_member] = key_b64_encoded
+    params[sse_md5_member] = key_md5_str
 
 
-def _needs_s3_sse_customization(params):
-    return (params.get('SSECustomerKey') is not None and
-            'SSECustomerKeyMD5' not in params)
-
-
-def check_dns_name(bucket_name):
-    """
-    Check to see if the ``bucket_name`` complies with the
-    restricted DNS naming conventions necessary to allow
-    access via virtual-hosting style.
-
-    Even though "." characters are perfectly valid in this DNS
-    naming scheme, we are going to punt on any name containing a
-    "." character because these will cause SSL cert validation
-    problems if we try to use virtual-hosting style addressing.
-    """
-    if '.' in bucket_name:
-        return False
-    n = len(bucket_name)
-    if n < 3 or n > 63:
-        # Wrong length
-        return False
-    if n == 1:
-        if not bucket_name.isalnum():
-            return False
-    match = LABEL_RE.match(bucket_name)
-    if match is None or match.end() != len(bucket_name):
-        return False
-    return True
-
-
-def fix_s3_host(request, signature_version, region_name, **kwargs):
-    """
-    This handler looks at S3 requests just before they are signed.
-    If there is a bucket name on the path (true for everything except
-    ListAllBuckets) it checks to see if that bucket name conforms to
-    the DNS naming conventions.  If it does, it alters the request to
-    use ``virtual hosting`` style addressing rather than ``path-style``
-    addressing.  This allows us to avoid 301 redirects for all
-    bucket names that can be CNAME'd.
-    """
-    if request.auth_path is not None:
-        # The auth_path has already been applied (this may be a
-        # retried request).  We don't need to perform this
-        # customization again.
-        return
-    elif _is_get_bucket_location_request(request):
-        # For the GetBucketLocation response, we should not be using
-        # the virtual host style addressing so we can avoid any sigv4
-        # issues.
-        logger.debug("Request is GetBucketLocation operation, not checking "
-                     "for DNS compatibility.")
-        return
-    parts = urlsplit(request.url)
-    request.auth_path = parts.path
-    path_parts = parts.path.split('/')
-    if signature_version in ['s3v4', 'v4']:
-        return
-    if len(path_parts) > 1:
-        bucket_name = path_parts[1]
-        logger.debug('Checking for DNS compatible bucket for: %s',
-                     request.url)
-        if check_dns_name(bucket_name) and _allowed_region(region_name):
-            # If the operation is on a bucket, the auth_path must be
-            # terminated with a '/' character.
-            if len(path_parts) == 2:
-                if request.auth_path[-1] != '/':
-                    request.auth_path += '/'
-            path_parts.remove(bucket_name)
-            global_endpoint = 's3.amazonaws.com'
-            host = bucket_name + '.' + global_endpoint
-            new_tuple = (parts.scheme, host, '/'.join(path_parts),
-                         parts.query, '')
-            new_uri = urlunsplit(new_tuple)
-            request.url = new_uri
-            logger.debug('URI updated to: %s', new_uri)
-        else:
-            logger.debug('Not changing URI, bucket is not DNS compatible: %s',
-                         bucket_name)
-
-
-def _is_get_bucket_location_request(request):
-    return request.url.endswith('?location')
-
-
-def _allowed_region(region_name):
-    return region_name not in RESTRICTED_REGIONS
+def _needs_s3_sse_customization(params, sse_member_prefix):
+    return (params.get(sse_member_prefix + 'Key') is not None and
+            sse_member_prefix + 'KeyMD5' not in params)
 
 
 def register_retries_for_service(service_data, session,
@@ -287,14 +268,86 @@ def add_expect_header(model, params, **kwargs):
             params['headers']['Expect'] = '100-continue'
 
 
-def quote_source_header(params, **kwargs):
-    if params['headers'] and 'x-amz-copy-source' in params['headers']:
-        value = params['headers']['x-amz-copy-source']
-        params['headers']['x-amz-copy-source'] = quote(
-            value.encode('utf-8'), '/~')
+def document_copy_source_form(section, event_name, **kwargs):
+    if 'request-example' in event_name:
+        parent = section.get_section('structure-value')
+        param_line = parent.get_section('CopySource')
+        value_portion = param_line.get_section('member-value')
+        value_portion.clear_text()
+        value_portion.write("'string' or {'Bucket': 'string', "
+                            "'Key': 'string', 'VersionId': 'string'}")
+    elif 'request-params' in event_name:
+        param_section = section.get_section('CopySource')
+        type_section = param_section.get_section('param-type')
+        type_section.clear_text()
+        type_section.write(':type CopySource: str or dict')
+        doc_section = param_section.get_section('param-documentation')
+        doc_section.clear_text()
+        doc_section.write(
+            "The name of the source bucket, key name of the source object, "
+            "and optional version ID of the source object.  You can either "
+            "provide this value as a string or a dictionary.  The "
+            "string form is {bucket}/{key} or "
+            "{bucket}/{key}?versionId={versionId} if you want to copy a "
+            "specific version.  You can also provide this value as a "
+            "dictionary.  The dictionary format is recommended over "
+            "the string format because it is more explicit.  The dictionary "
+            "format is: {'Bucket': 'bucket', 'Key': 'key', 'VersionId': 'id'}."
+            "  Note that the VersionId key is optional and may be omitted."
+        )
 
 
-def copy_snapshot_encrypted(params, request_signer, endpoint, **kwargs):
+def handle_copy_source_param(params, **kwargs):
+    """Convert CopySource param for CopyObject/UploadPartCopy.
+
+    This handler will deal with two cases:
+
+        * CopySource provided as a string.  We'll make a best effort
+          to URL encode the key name as required.  This will require
+          parsing the bucket and version id from the CopySource value
+          and only encoding the key.
+        * CopySource provided as a dict.  In this case we're
+          explicitly given the Bucket, Key, and VersionId so we're
+          able to encode the key and ensure this value is serialized
+          and correctly sent to S3.
+
+    """
+    source = params.get('CopySource')
+    if source is None:
+        # The call will eventually fail but we'll let the
+        # param validator take care of this.  It will
+        # give a better error message.
+        return
+    if isinstance(source, six.string_types):
+        params['CopySource'] = _quote_source_header(source)
+    elif isinstance(source, dict):
+        params['CopySource'] = _quote_source_header_from_dict(source)
+
+
+def _quote_source_header_from_dict(source_dict):
+    try:
+        bucket = source_dict['Bucket']
+        key = percent_encode(source_dict['Key'], safe=SAFE_CHARS + '/')
+        version_id = source_dict.get('VersionId')
+    except KeyError as e:
+        raise ParamValidationError(
+            report='Missing required parameter: %s' % str(e))
+    final =  '%s/%s' % (bucket, key)
+    if version_id is not None:
+        final += '?versionId=%s' % version_id
+    return final
+
+
+def _quote_source_header(value):
+    result = VERSION_ID_SUFFIX.search(value)
+    if result is None:
+        return percent_encode(value, safe=SAFE_CHARS + '/')
+    else:
+        first, version_id = value[:result.start()], value[result.start():]
+        return percent_encode(first, safe=SAFE_CHARS + '/') + version_id
+
+
+def copy_snapshot_encrypted(params, request_signer, **kwargs):
     # The presigned URL that facilities copying an encrypted snapshot.
     # If the user does not provide this value, we will automatically
     # calculate on behalf of the user and inject the PresignedUrl
@@ -314,11 +367,6 @@ def copy_snapshot_encrypted(params, request_signer, endpoint, **kwargs):
     # url based on the source endpoint.
     source_region = params['SourceRegion']
 
-    presigner = request_signer.get_auth(
-        'ec2', source_region, signature_version='v4-query')
-
-    request = endpoint.create_request(request_dict)
-    original = request.original
     # The better way to do this is to actually get the
     # endpoint_resolver and get the endpoint_url given the
     # source region.  In this specific case, we know that
@@ -327,11 +375,14 @@ def copy_snapshot_encrypted(params, request_signer, endpoint, **kwargs):
     # general this is not a safe assumption to make.
     # I think eventually we should try to plumb through something
     # that allows us to resolve endpoints from regions.
-    original.url = original.url.replace(destination_region, source_region)
-
-    presigner.add_auth(request=original)
-    request = original.prepare()
-    params['PresignedUrl'] = original.url
+    request_dict_copy = copy.deepcopy(request_dict)
+    request_dict_copy['url'] = request_dict['url'].replace(
+        destination_region, source_region)
+    request_dict_copy['method'] = 'GET'
+    request_dict_copy['headers'] = {}
+    presigned_url = request_signer.generate_presigned_url(
+        request_dict_copy, region_name=source_region)
+    params['PresignedUrl'] = presigned_url
 
 
 def json_decode_policies(parsed, model, **kwargs):
@@ -355,7 +406,8 @@ def _decode_policy_types(parsed, shape):
             if member_shape.type_name == 'string' and \
                     member_shape.name == shape_name and \
                     member_name in parsed:
-                parsed[member_name] = decode_quoted_jsondoc(parsed[member_name])
+                parsed[member_name] = decode_quoted_jsondoc(
+                    parsed[member_name])
             elif member_name in parsed:
                 _decode_policy_types(parsed[member_name], member_shape)
     if shape.type_name == 'list':
@@ -389,6 +441,14 @@ def base64_encode_user_data(params, **kwargs):
             params['UserData']).decode('utf-8')
 
 
+def document_base64_encoding():
+    description = ('**This value will be base64 encoded automatically. Do '
+                   'not base64 encode this value prior to performing the '
+                   'operation.**')
+    append = AppendParamDocumentation('UserData', description)
+    return append.append_documentation
+
+
 def fix_route53_ids(params, model, **kwargs):
     """
     Check for and split apart Route53 resource IDs, setting
@@ -421,8 +481,13 @@ def inject_account_id(params, **kwargs):
 
 def add_glacier_version(model, params, **kwargs):
     request_dict = params
-    request_dict['headers']['x-amz-glacier-version'] = \
-            model.metadata['apiVersion']
+    request_dict['headers']['x-amz-glacier-version'] = model.metadata[
+        'apiVersion']
+
+
+def add_accept_header(model, params, **kwargs):
+    request_dict = params
+    request_dict['headers']['Accept'] = 'application/json'
 
 
 def add_glacier_checksums(params, **kwargs):
@@ -457,12 +522,29 @@ def add_glacier_checksums(params, **kwargs):
     body.seek(starting_position)
 
 
+def document_glacier_tree_hash_checksum():
+    doc = '''
+        This is a required field.
+
+        Ideally you will want to compute this value with checksums from
+        previous uploaded parts, using the algorithm described in
+        `Glacier documentation <http://docs.aws.amazon.com/amazonglacier/latest/dev/checksum-calculations.html>`_.
+
+        But if you prefer, you can also use botocore.util.calculate_tree_hash()
+        to compute it from raw file by::
+
+            checksum = calculate_tree_hash(open('your_file.txt', 'rb'))
+
+        '''
+    return AppendParamDocumentation('checksum', doc).append_documentation
+
+
 def switch_host_machinelearning(request, **kwargs):
     switch_host_with_param(request, 'PredictEndpoint')
 
 
 def switch_host_with_param(request, param_name):
-    request_json = json.loads(request.data)
+    request_json = json.loads(request.data.decode('utf-8'))
     if request_json.get(param_name):
         new_endpoint = request_json[param_name]
         new_endpoint_components = urlsplit(new_endpoint)
@@ -479,26 +561,96 @@ def switch_host_with_param(request, param_name):
         request.url = final_endpoint
 
 
+def check_openssl_supports_tls_version_1_2(**kwargs):
+    import ssl
+    try:
+        openssl_version_tuple = ssl.OPENSSL_VERSION_INFO
+        if openssl_version_tuple[0] < 1 or openssl_version_tuple[2] < 1:
+            warnings.warn(
+                'Currently installed openssl version: %s does not '
+                'support TLS 1.2, which is required for use of iot-data. '
+                'Please use python installed with openssl version 1.0.1 or '
+                'higher.' % (ssl.OPENSSL_VERSION),
+                UnsupportedTLSVersionWarning
+            )
+    # We cannot check the openssl version on python2.6, so we should just
+    # pass on this conveniency check.
+    except AttributeError:
+        pass
+
+
+def change_get_to_post(request, **kwargs):
+    # This is useful when we need to change a potentially large GET request
+    # into a POST with x-www-form-urlencoded encoding.
+    if request.method == 'GET' and '?' in request.url:
+        request.headers['Content-Type'] = 'application/x-www-form-urlencoded'
+        request.method = 'POST'
+        request.url, request.data = request.url.split('?', 1)
+
+
+def set_list_objects_encoding_type_url(params, context, **kwargs):
+    if 'EncodingType' not in params:
+        # We set this context so that we know it wasn't the customer that
+        # requested the encoding.
+        context['EncodingTypeAutoSet'] = True
+        params['EncodingType'] = 'url'
+
+
+def decode_list_object(parsed, context, **kwargs):
+    # This is needed because we are passing url as the encoding type. Since the
+    # paginator is based on the key, we need to handle it before it can be
+    # round tripped.
+    if 'Contents' in parsed and parsed.get('EncodingType') == 'url' and \
+                    context.get('EncodingTypeAutoSet'):
+        for content in parsed['Contents']:
+            content['Key'] = unquote_str(content['Key'])
+
+
 # This is a list of (event_name, handler).
 # When a Session is created, everything in this list will be
 # automatically registered with that Session.
 
 BUILTIN_HANDLERS = [
+    ('creating-client-class', add_generate_presigned_url),
+    ('creating-client-class.s3', add_generate_presigned_post),
+    ('creating-client-class.iot-data', check_openssl_supports_tls_version_1_2),
     ('after-call.iam', json_decode_policies),
 
     ('after-call.ec2.GetConsoleOutput', decode_console_output),
     ('after-call.cloudformation.GetTemplate', json_decode_template_body),
     ('after-call.s3.GetBucketLocation', parse_get_bucket_location),
 
+    ('before-parameter-build.s3', validate_bucket_name),
+
+    ('before-parameter-build.s3.ListObjects',
+     set_list_objects_encoding_type_url),
     ('before-call.s3.PutBucketTagging', calculate_md5),
     ('before-call.s3.PutBucketLifecycle', calculate_md5),
+    ('before-call.s3.PutBucketLifecycleConfiguration', calculate_md5),
     ('before-call.s3.PutBucketCors', calculate_md5),
     ('before-call.s3.DeleteObjects', calculate_md5),
     ('before-call.s3.PutBucketReplication', calculate_md5),
-    ('before-call.s3.UploadPartCopy', quote_source_header),
-    ('before-call.s3.CopyObject', quote_source_header),
+    ('before-call.s3.PutObject', conditionally_calculate_md5),
+    ('before-call.s3.UploadPart', conditionally_calculate_md5),
+    ('before-call.s3.PutBucketAcl', conditionally_calculate_md5),
+    ('before-call.s3.PutBucketLogging', conditionally_calculate_md5),
+    ('before-call.s3.PutBucketNotification', conditionally_calculate_md5),
+    ('before-call.s3.PutBucketPolicy', conditionally_calculate_md5),
+    ('before-call.s3.PutBucketRequestPayment', conditionally_calculate_md5),
+    ('before-call.s3.PutBucketVersioning', conditionally_calculate_md5),
+    ('before-call.s3.PutBucketWebsite', conditionally_calculate_md5),
+    ('before-call.s3.PutObjectAcl', conditionally_calculate_md5),
+
+    ('before-parameter-build.s3.CopyObject',
+     handle_copy_source_param),
+    ('before-parameter-build.s3.UploadPartCopy',
+     handle_copy_source_param),
+    ('docs.*.s3.CopyObject.complete-section', document_copy_source_form),
+    ('docs.*.s3.UploadPartCopy.complete-section', document_copy_source_form),
+
     ('before-call.s3', add_expect_header),
     ('before-call.glacier', add_glacier_version),
+    ('before-call.apigateway', add_accept_header),
     ('before-call.glacier.UploadArchive', add_glacier_checksums),
     ('before-call.glacier.UploadMultipartPart', add_glacier_checksums),
     ('before-call.ec2.CopySnapshot', copy_snapshot_encrypted),
@@ -510,18 +662,66 @@ BUILTIN_HANDLERS = [
     ('service-data-loaded', register_retries_for_service),
     ('choose-signer.cognito-identity.GetId', disable_signing),
     ('choose-signer.cognito-identity.GetOpenIdToken', disable_signing),
+    ('choose-signer.cognito-identity.UnlinkIdentity', disable_signing),
+    ('choose-signer.cognito-identity.GetCredentialsForIdentity',
+        disable_signing),
     ('choose-signer.sts.AssumeRoleWithSAML', disable_signing),
-    ('before-sign.s3', fix_s3_host),
+    ('choose-signer.sts.AssumeRoleWithWebIdentity', disable_signing),
+    ('before-sign.s3', utils.fix_s3_host),
     ('before-parameter-build.s3.HeadObject', sse_md5),
     ('before-parameter-build.s3.GetObject', sse_md5),
     ('before-parameter-build.s3.PutObject', sse_md5),
     ('before-parameter-build.s3.CopyObject', sse_md5),
+    ('before-parameter-build.s3.CopyObject', copy_source_sse_md5),
     ('before-parameter-build.s3.CreateMultipartUpload', sse_md5),
     ('before-parameter-build.s3.UploadPart', sse_md5),
     ('before-parameter-build.s3.UploadPartCopy', sse_md5),
+    ('before-parameter-build.s3.UploadPartCopy', copy_source_sse_md5),
     ('before-parameter-build.ec2.RunInstances', base64_encode_user_data),
     ('before-parameter-build.autoscaling.CreateLaunchConfiguration',
      base64_encode_user_data),
     ('before-parameter-build.route53', fix_route53_ids),
     ('before-parameter-build.glacier', inject_account_id),
+    ('after-call.s3.ListObjects', decode_list_object),
+
+    # Cloudsearchdomain search operation will be sent by HTTP POST
+    ('request-created.cloudsearchdomain.Search',
+     change_get_to_post),
+    # Glacier documentation customizations
+    ('docs.*.glacier.*.complete-section',
+     AutoPopulatedParam('accountId', 'Note: this parameter is set to "-" by \
+                         default if no value is not specified.')
+     .document_auto_populated_param),
+    ('docs.*.glacier.UploadArchive.complete-section',
+     AutoPopulatedParam('checksum').document_auto_populated_param),
+    ('docs.*.glacier.UploadMultipartPart.complete-section',
+     AutoPopulatedParam('checksum').document_auto_populated_param),
+    ('docs.request-params.glacier.CompleteMultipartUpload.complete-section',
+     document_glacier_tree_hash_checksum()),
+
+    # UserData base64 encoding documentation customizations
+    ('docs.*.ec2.RunInstances.complete-section', document_base64_encoding()),
+    ('docs.*.autoscaling.CreateLaunchConfiguration.complete-section',
+     document_base64_encoding()),
+    # EC2 CopySnapshot documentation customizations
+    ('docs.*.ec2.CopySnapshot.complete-section',
+     AutoPopulatedParam('PresignedUrl').document_auto_populated_param),
+    ('docs.*.ec2.CopySnapshot.complete-section',
+     AutoPopulatedParam('DestinationRegion').document_auto_populated_param),
+    # S3 SSE documentation modifications
+    ('docs.*.s3.*.complete-section',
+     AutoPopulatedParam('SSECustomerKeyMD5').document_auto_populated_param),
+    # S3 SSE Copy Source documentation modifications
+    ('docs.*.s3.*.complete-section',
+     AutoPopulatedParam(
+        'CopySourceSSECustomerKeyMD5').document_auto_populated_param),
+    # The following S3 operations cannot actually accept a ContentMD5
+    ('docs.*.s3.*.complete-section',
+     HideParamFromOperations(
+         's3', 'ContentMD5',
+         ['DeleteObjects', 'PutBucketAcl', 'PutBucketCors',
+          'PutBucketLifecycle', 'PutBucketLogging', 'PutBucketNotification',
+          'PutBucketPolicy', 'PutBucketReplication', 'PutBucketRequestPayment',
+          'PutBucketTagging', 'PutBucketVersioning', 'PutBucketWebsite',
+          'PutObjectAcl']).hide_param)
 ]
