@@ -95,11 +95,10 @@ import base64
 import json
 import xml.etree.cElementTree
 import logging
-from pprint import pformat
 
 from botocore.compat import six, XMLParseError
 
-from botocore.utils import parse_timestamp
+from botocore.utils import parse_timestamp, merge_dicts
 
 LOG = logging.getLogger(__name__)
 
@@ -201,7 +200,7 @@ class ResponseParser(object):
             always be present.
 
         """
-        LOG.debug('Response headers:\n%s', pformat(dict(response['headers'])))
+        LOG.debug('Response headers: %s', response['headers'])
         LOG.debug('Response body:\n%s', response['body'])
         if response['status_code'] >= 301:
             parsed = self._do_error_parse(response, shape)
@@ -312,6 +311,11 @@ class BaseXMLResponseParser(ResponseParser):
         return member_name
 
     def _build_name_to_xml_node(self, parent_node):
+        # If the parent node is actually a list. We should not be trying
+        # to serialize it to a dictionary. Instead, return the first element
+        # in the list.
+        if isinstance(parent_node, list):
+            return self._build_name_to_xml_node(parent_node[0])
         xml_dict = {}
         for item in parent_node:
             key = self._node_tag(item)
@@ -461,6 +465,11 @@ class BaseJSONParser(ResponseParser):
 
     def _handle_structure(self, shape, value):
         member_shapes = shape.members
+        if value is None:
+            # If the comes across the wire as "null" (None in python),
+            # we should be returning this unchanged, instead of as an
+            # empty dict.
+            return None
         final_parsed = {}
         for member_name in member_shapes:
             member_shape = member_shapes[member_name]
@@ -488,21 +497,6 @@ class BaseJSONParser(ResponseParser):
     def _handle_timestamp(self, shape, value):
         return self._timestamp_parser(value)
 
-
-class JSONParser(BaseJSONParser):
-    """Response parse for the "json" protocol."""
-    def _do_parse(self, response, shape):
-        # The json.loads() gives us the primitive JSON types,
-        # but we need to traverse the parsed JSON data to convert
-        # to richer types (blobs, timestamps, etc.
-        parsed = {}
-        if shape is not None:
-            body = response['body'].decode(self.DEFAULT_ENCODING)
-            original_parsed = json.loads(body)
-            parsed = self._parse_shape(shape, original_parsed)
-        self._inject_response_metadata(parsed, response['headers'])
-        return parsed
-
     def _do_error_parse(self, response, shape):
         body = json.loads(response['body'].decode(self.DEFAULT_ENCODING))
         error = {"Error": {}, "ResponseMetadata": {}}
@@ -533,6 +527,21 @@ class JSONParser(BaseJSONParser):
                 headers['x-amzn-requestid'])
 
 
+class JSONParser(BaseJSONParser):
+    """Response parse for the "json" protocol."""
+    def _do_parse(self, response, shape):
+        # The json.loads() gives us the primitive JSON types,
+        # but we need to traverse the parsed JSON data to convert
+        # to richer types (blobs, timestamps, etc.
+        parsed = {}
+        if shape is not None:
+            body = response['body'].decode(self.DEFAULT_ENCODING)
+            original_parsed = json.loads(body)
+            parsed = self._parse_shape(shape, original_parsed)
+        self._inject_response_metadata(parsed, response['headers'])
+        return parsed
+
+
 class BaseRestParser(ResponseParser):
 
     def _do_parse(self, response, shape):
@@ -633,7 +642,7 @@ class RestJSONParser(BaseRestParser, BaseJSONParser):
 
     def _do_error_parse(self, response, shape):
         body = self._initial_body_parse(response['body'])
-        error = {'Error': {}, 'ResponseMetadata': {}}
+        error = super(RestJSONParser, self)._do_error_parse(response, shape)
         error['Error']['Message'] = body.get('message',
                                              body.get('Message', ''))
         if 'x-amzn-errortype' in response['headers']:
@@ -670,9 +679,16 @@ class RestXMLParser(BaseRestParser, BaseXMLResponseParser):
         #   <RequestId>request-id</RequestId>
         # </ErrorResponse>
         if response['body']:
-            return self._parse_error_from_body(response)
-        else:
-            return self._parse_error_from_http_status(response)
+            # If the body ends up being invalid xml, the xml parser should not
+            # blow up. It should at least try to pull information about the
+            # the error response from other sources like the HTTP status code.
+            try:
+                return self._parse_error_from_body(response)
+            except ResponseParserError as e:
+                LOG.debug(
+                    'Exception caught when parsing error response body:',
+                    exc_info=True)
+        return self._parse_error_from_http_status(response)
 
     def _parse_error_from_http_status(self, response):
         return {
@@ -706,7 +722,9 @@ class RestXMLParser(BaseRestParser, BaseXMLResponseParser):
         elif 'RequestId' in parsed:
             # Other rest-xml serivces:
             parsed['ResponseMetadata'] = {'RequestId': parsed.pop('RequestId')}
-        return parsed
+        default = {'Error': {'Message': '', 'Code': ''}}
+        merge_dicts(default, parsed)
+        return default
 
 
 PROTOCOL_PARSERS = {
